import inspect
import threading
import copy
from functools import wraps
import logging
from .circular_reference import circular_reference
from typing import Tuple, Callable, ParamSpec, TypeVar, Any
from collections import OrderedDict
from .cache_exception import CacheException

P = ParamSpec("P")
T = TypeVar("T")


class Memoiz:

    def __init__(
        self,
        iterables: Tuple[type, ...] = (list, tuple, set),
        mappables: Tuple[type, ...] = (dict, OrderedDict),
        sortables: Tuple[type, ...] = (dict, set),
        deep_copy: bool = True,
        *args,
        **kwargs,
    ):
        self.deep_copy = deep_copy
        self.mappables = mappables
        self.iterables = iterables
        self.sortables = sortables
        self._cache = {}
        self._lock = threading.Lock()

    def clear(self, callable: Callable, *args, **kwargs) -> None:
        with self._lock:
            args_key = self._freeze((args, kwargs))
            del self._cache[callable][args_key]
            if len(self._cache[callable]) == 0:
                del self._cache[callable]

    def clear_callable(self, callable: Callable) -> None:
        with self._lock:
            del self._cache[callable]

    def clear_all(self) -> None:
        self._cache = {}

    def _freeze(self, it: Any, seen: list = None) -> Any:
        if seen is None:
            seen = []
        try:
            hash(it)
            return it
        except Exception as e:
            pass
        if isinstance(it, self.iterables):
            if any(it is i for i in seen):
                return circular_reference
            seen.append(it)
            if type(it) in self.sortables:
                it = sorted(it, key=str)
            return tuple(self._freeze(i, seen) for i in it)
        elif isinstance(it, self.mappables):
            if any(it is i for i in seen):
                return circular_reference
            seen.append(it)
            if type(it) in self.sortables:
                its = sorted(it.items(), key=lambda x: str(x[0]))
            else:
                its = it.items()
            return tuple((k, self._freeze(v, seen)) for k, v in its)

        raise CacheException(f"Cannot freeze {it}.")

    def __call__(self, callable: Callable[P, T]) -> Callable[P, T]:
        @wraps(callable)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            try:
                if len(args) != 0 and (
                    hasattr(args[0], callable.__name__)
                    and inspect.unwrap(getattr(args[0], callable.__name__)) is callable
                ):
                    # If the first argument is an object and it contains the method `callable` then use the unwrapped method (i.e., the bound function) for the key.
                    # This is necessary because the bound function is the reference that may be used for clearing a chache entry.
                    callable_key = getattr(args[0], callable.__name__)
                    args_key = self._freeze((args[1:], kwargs))
                else:
                    # If this is not a method call, then use the wrapper for the key.  This is necessary, as referencing the function will return the wrapper.
                    callable_key = wrapper
                    args_key = self._freeze((args, kwargs))

                if callable_key in self._cache and args_key in self._cache[callable_key]:
                    logging.debug(f"Using cache for {(callable_key, args_key)}.")
                    self._lock.acquire()
                    result = self._cache[callable_key][args_key]
                    self._lock.release()
                else:
                    result = callable(*args, **kwargs)
                    with self._lock:
                        if callable_key not in self._cache:
                            self._cache[callable_key] = {}
                        if args_key not in self._cache[callable_key]:
                            self._cache[callable_key][args_key] = result
                            logging.debug(f"Cached {(callable_key, args_key)}.")

                if self.deep_copy:
                    return copy.deepcopy(result)
                else:
                    return result
            except CacheException as e:
                logging.debug(e)
                return callable(*args, **kwargs)
            except BaseException as e:
                if self._lock.locked():
                    self._lock.release()
                raise e

        return wrapper



import unittest
from typing import Any
from collections import OrderedDict
from memoiz import Memoiz, circular_reference


class TestCase(unittest.TestCase):

    def setUp(self) -> None:

        cache = Memoiz()

        class Test:

            @cache
            def identity(self, *args, **kwargs) -> Any:
                return args, kwargs

        @cache
        def identity(*args, **kwargs) -> Any:
            return args, kwargs

        @cache
        def callable(arg0: Any) -> Any:
            return identity(arg0)

        self.cache = cache
        self.test = Test()
        self.identity = identity
        self.callable = callable

    def test_cache_entry(self) -> None:
        self.identity({"a": 42})
        self.assertIn(self.identity, self.cache._cache)

    def test_cache_callable_entry(self) -> None:
        self.identity({"a": 42})
        self.assertIn((((("a", 42),),), ()), self.cache._cache[self.identity])

    def test_callstack_deadlock(self) -> None:
        result = self.callable(42)
        self.assertEqual(((42,), {}), result)

    def test_cicular_reference(self) -> None:
        x = []
        x.append(x)
        self.identity(x)
        self.assertEqual((([x],), {}), self.cache._cache[self.identity][(((circular_reference,),), ())])

    def test_removal_of_entry_from_cache_for_function_with_args(self) -> None:
        self.identity({"a": 42}, a=42)
        self.identity({"a": 23}, a=23)
        self.cache.clear(self.identity, {"a": 42}, a=42)
        self.assertEqual(
            {self.identity: {(((("a", 23),),), ((("a", 23),))): (({"a": 23},), {"a": 23})}}, self.cache._cache
        )

    def test_removal_of_function_from_cache_with_args(self) -> None:
        self.identity({"a": 42}, a=42)
        self.cache.clear(self.identity, {"a": 42}, a=42)
        self.assertEqual({}, self.cache._cache)

    def test_removal_of_function_from_cache(self) -> None:
        self.identity({"a": 42})
        self.cache.clear_callable(self.identity)
        self.assertEqual({}, self.cache._cache)

    def test_removal_of_entry_for_method(self) -> None:
        self.test.identity({"a": 42}, a=42)
        self.test.identity({"a": 23}, a=23)
        self.cache.clear(self.test.identity, {"a": 42}, a=42)
        self.assertEqual(
            {self.test.identity: {(((("a", 23),),), (("a", 23),)): (({"a": 23},), {"a": 23})}}, self.cache._cache
        )

    def test_removal_of_method_with_args(self) -> None:
        self.test.identity({"a": 42})
        self.cache.clear(self.test.identity, {"a": 42})
        self.assertEqual({}, self.cache._cache)

    def test_deterministic_representation_of_sets(self) -> None:
        for _ in range(0, int(1e4)):
            s1 = {23, "42", 57, ...}
            s2 = {..., 23, "42", 57}
            self.assertEqual(s1, s2)
            self.identity(s1)
            self.identity(s2)
            self.assertEqual(1, len(self.cache._cache[self.identity]))

    def test_deterministic_representation_of_dicts(self) -> None:
        for _ in range(0, int(1e4)):
            d1 = {...: 57, "a": 23, "b": 42}
            d2 = {"b": 42, "a": 23, ...: 57}
            self.assertEqual(d1, d2)
            self.identity(d1)
            self.identity(d2)
            self.assertEqual(1, len(self.cache._cache[self.identity]))

    def test_insertion_order_of_ordereddicts(self) -> None:
        d1 = OrderedDict(a=23, b=42, c=57)
        d2 = OrderedDict(c=57, b=42, a=23)
        d3 = OrderedDict(c=57, b=42, a=23)
        self.assertNotEqual(d1, d2)
        self.identity(d1)
        self.identity(d2)
        self.identity(d3)
        self.assertEqual(2, len(self.cache._cache[self.identity]))


if __name__ == "__main__":
    unittest.main()
